#include "eq_table.h"
#include "expr_parser.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ltdl.h>

#if 0
#define OUTPUT
#else
#undef OUTPUT
#endif

#define Q_PLUS_Q 0
#define Q_PLUS_S 1
#define Z_PLUS_Z 2
#define Z_PLUS_S 3
#define S_PLUS_S 4

#define Q_MINUS_Q 5
#define Q_MINUS_S 6
#define S_MINUS_Q 7
#define Z_MINUS_Z 8
#define Z_MINUS_S 9
#define S_MINUS_Z 10
#define S_MINUS_S 11

#define Q_TIMES_Q 12
#define Q_TIMES_Q_NC 13
#define Q_TIMES_S 14
#define Z_TIMES_Z 15
#define Z_TIMES_S 16
#define S_TIMES_S 17

#define Q_DIV_S	 18
#define S_DIV_S	 19
#define Q_DIV_Q	 20
#define S_DIV_Q	 21
#define Z_DIV_S	 22
#define Z_DIV_Z	 23
#define S_DIV_Z	 24

#define Q_EXP_S_EVEN 25
#define S_EXP_Q	26
#define Z_EXP_S_EVEN 27
#define S_EXP_Z	28
#define S_EXP_S 29

#define RE_Q    30
#define IM_I_Q	31
#define IM_J_Q	32
#define IM_K_Q	33

#define ABS_S     34
#define ABS_Q     35
#define ABS_Z     36

#define SIN_Z   37
#define COS_Z   38
#define TAN_Z   39

#define SINH_Z   40
#define COSH_Z   41
#define TANH_Z   42

#define LN_Z  43
#define LOG_Z 44

extern char * yy_scan_string(const char *);
extern expr_ref tree_top;
extern int yyparse();

eq_table::eq_table(eq_table *copy) {

    lt_dlinit();		// init ltdl

    max_iter = copy->max_iter;
    max_val = copy->max_val;
    num_iterates = copy->num_iterates;
    index_of_last_column = copy->index_of_last_column;
    c_index = copy->c_index;
    q_index = copy->q_index;
    table = new Entry*[index_of_last_column+1];
    for(int i = 0 ; i <= index_of_last_column ; i++) {
	table[i] = new Entry(copy->table[i]);
    }
    init_table = new Entry*[init_index_of_last_column+1];
    for(int i = 0 ; i <= init_index_of_last_column ; i++) {
	init_table[i] = new Entry(copy->init_table[i]);
    }
    deal_with_old_iterates();
}

eq_table::Entry::Entry(Entry * copy) {
    
    for(int i = 0 ; i < 4 ; i++) {
	vect[i] = copy->vect[i];
    }
    index = copy->index;
    expr = copy->expr;
}
				// this function works but is no longer
				// necessary :( 
void eq_table::Entry::swap(Entry** front) {
				// moves entry before current to after
    Entry * a, *b, *c, *d;	// current is c

    if(prev) {
	b = prev;
	c = this;
	printf("swapping %d and %d\n", b->index, c->index);
	b->index++;
	c->index--;
	if(prev->prev) {
	    if(next) {
		a = prev->prev;	// --a--b--c--d--
		d = next;	// --a--c--b--d--
				
		a->next = c;
		b->next = d;
		c->next = b;
		d->prev = b;
		b->prev = c;
		c->prev = a;
	    } else {
		a = prev->prev;	// --a--b--c--0
		d = NULL;	// --a--c--b--0
				
		a->next = c;
		b->next = d;
		c->next = b;
		b->prev = c;
		c->prev = a;
	    }
	} else {
	    if(next) {
		a = NULL;	// 0--b--c--d--
		d = next;	// 0--c--b--d-- 
		
		*front = c;
		b->next = d;
		c->next = b;
		d->prev = b;
		b->prev = c;
		c->prev = a;
	    } else {
		a = NULL;	// 0--b--c--0
		d = NULL;	// 0--c--b--0
		    
		*front = c;
		b->next = d;
		c->next = b;
		b->prev = c;
		c->prev = a;
	    }
	}
    }
}

void eq_table::add_expr(expr_ref expr, Entry ** list) {

    Entry * ptr = add_column(expr,list);

//    printf("data is %s\n", expr->data);
//    printf("ptr is %p\n", ptr);

    expr_ref e;
    int val;

//    printf("ptr is %p\n", ptr);

    switch (expr->token_code) {

    case '|':
	if(expr->scalar)
	    ptr->rule = ABS_S;
	else 
	    ptr->rule = ABS_Q;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_RE:
	ptr->rule = RE_Q;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_IM_I:
	ptr->rule = IM_I_Q;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_IM_J:
	ptr->rule = IM_J_Q;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_IM_K:
	ptr->rule = IM_K_Q;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_SIN:
	ptr->rule = SIN_Z;
      	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_COS:
	ptr->rule = COS_Z;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_TAN:
	ptr->rule = TAN_Z;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_SINH:
	ptr->rule = SINH_Z;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_COSH:
	ptr->rule = COSH_Z;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_TANH:
	ptr->rule = TANH_Z;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_LN:
	ptr->rule = LN_Z;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = -1;
	break;

    case TOK_LOG:
	ptr->rule = LOG_Z;
	ptr->dep1 = get_column(expr->rchld,list);
	ptr->dep2 = get_column(expr->lchld,list);
	break;
	
    case '_':		// subscript
	e = expr;
	while(e->rchld) e = e->rchld;
	    
	val = atoi(e->data);
	if(!e->token_code == TOK_LIT_SCALAR) {
	    fprintf(stderr, "scalar not found in subscript\n");
	    exit(-1);
	}
	if(num_iterates < val + 1)
	    num_iterates = val + 1;
	break;

    case 'c':
	if(*list == table_list)	// this is a hack ;<
	    c_index = ptr->index;
	else
	    init_c_index = ptr->index;
	break;
    case 'q':
	if(expr->token_code == 'q')
	    q_index = ptr->index;
	break;
    case 'z':
	if(expr->token_code == 'z')
	    q_index = ptr->index;
	break;
    case 'i':
    case 'j':
    case 'k':
    case TOK_LIT_QUAT:
	break;
		
    case 'e':
    case TOK_LIT_PI:
    case TOK_LIT_SCALAR:
	break;

    case '+':
	if(expr->lchld->scalar) {
	    if(expr->rchld->scalar) {
		ptr->rule = S_PLUS_S;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    } else if(expr->rchld->complex) {
		ptr->rule = Z_PLUS_S;
		ptr->dep1 = get_column(expr->rchld,list);
		ptr->dep2 = get_column(expr->lchld,list);
	    } else {
		ptr->rule = Q_PLUS_S;
		ptr->dep1 = get_column(expr->rchld,list);
		ptr->dep2 = get_column(expr->lchld,list);
	    }
	} else if(expr->lchld->complex) {
	    if(expr->rchld->scalar) {
		ptr->rule = Z_PLUS_S;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    } else if(expr->rchld->complex) {
		ptr->rule = Z_PLUS_Z;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    } else {
		ptr->rule = Q_PLUS_Q;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    }
	} else {
	    if(expr->rchld->scalar) {
		ptr->rule = Q_PLUS_S;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    } else {
		ptr->rule = Q_PLUS_Q;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    }
	}
	break;

    case '-':
	ptr->dep1 = get_column(expr->lchld,list);
	ptr->dep2 = get_column(expr->rchld,list);

	if(expr->lchld->scalar) 
	    if(expr->rchld->scalar)
		ptr->rule = S_MINUS_S;
	    else if(expr->rchld->complex)
		ptr->rule = S_MINUS_Z;
	    else 
		ptr->rule = S_MINUS_Q;
	
	else if(expr->lchld->complex) 
	    if(expr->rchld->scalar) 
		ptr->rule = Z_MINUS_S;
	    else if(expr->rchld->complex) 
		ptr->rule = Z_MINUS_Z;
	    else 
		ptr->rule = Q_MINUS_Q;
	else 
	    if(expr->rchld->scalar)
		ptr->rule = Q_MINUS_S;
	    else 
		ptr->rule = Q_MINUS_Q;
	break;
				
    case '*':
	if(expr->lchld->scalar) {
	    if(expr->rchld->scalar) {
		ptr->rule = S_TIMES_S;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    } else if(expr->rchld->complex) {
		ptr->rule = Z_TIMES_S;
		ptr->dep1 = get_column(expr->rchld,list);
		ptr->dep2 = get_column(expr->lchld,list);
	    } else {
		ptr->rule = Q_TIMES_S;
		ptr->dep1 = get_column(expr->rchld,list);
		ptr->dep2 = get_column(expr->lchld,list);
	    }
	} else if(expr->lchld->complex) {
	    if(expr->rchld->scalar) {
		ptr->rule = Z_TIMES_S;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    } else if(expr->rchld->complex) {
		ptr->rule = Z_TIMES_Z;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    } else {
		if(ptr->expr->cross_product_drops_out)
		    ptr->rule = Q_TIMES_Q_NC;
		else
		    ptr->rule = Q_TIMES_Q;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    }
	} else {
	    if(expr->rchld->scalar) {
		ptr->rule = Q_TIMES_S;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    } else {
		if(ptr->expr->cross_product_drops_out)
		    ptr->rule = Q_TIMES_Q_NC;
		else
		    ptr->rule = Q_TIMES_Q;
		ptr->dep1 = get_column(expr->lchld,list);
		ptr->dep2 = get_column(expr->rchld,list);
	    }
	}
	break;

    case '/':
	if(expr->rchld->scalar) 
	    if(expr->lchld->scalar)
		ptr->rule = S_DIV_S;
	    else if(expr->lchld->complex) 
		ptr->rule = Z_DIV_S;
	    else
		ptr->rule = Q_DIV_S;
	else if(expr->rchld->complex)
	    if(expr->lchld->scalar)
		ptr->rule = S_DIV_Z;
	    else if(expr->lchld->complex)
		ptr->rule = Z_DIV_Z;
	    else
		ptr->rule = Q_DIV_Q;
	else
	    if(expr->lchld->scalar)
		ptr->rule = S_DIV_Q;
	    else
		ptr->rule = Q_DIV_Q;

	ptr->dep1 = get_column(expr->lchld,list);
	ptr->dep2 = get_column(expr->rchld,list);
	break;

    case '^':
	if(expr->lchld->scalar) { // scalar base
	    if(expr->rchld->scalar)
		ptr->rule = S_EXP_S;
	    else if(expr->rchld->complex)
		ptr->rule = S_EXP_Z;
	    else
		ptr->rule = S_EXP_Q;
	    ptr->dep1 = get_column(expr->lchld,list);
	    ptr->dep2 = get_column(expr->rchld,list);
	} else if(expr->lchld->complex) {
	    ptr->rule = Z_EXP_S_EVEN;
	    ptr->dep2 = -1;
	    if(expr->rchld->val->s == 2) {
		ptr->dep1 = get_column(expr->lchld,list);
	    } else {
		fprintf(stderr, "libquat: exponentiation error\n");
		exit(-1);
	    }
	} else {		// quaternion base
	    ptr->rule = Q_EXP_S_EVEN;
	    ptr->dep2 = -1;	
	    if(expr->rchld->val->s == 2) {
		ptr->dep1 = get_column(expr->lchld,list);
	    } else {
		fprintf(stderr, "libquat: exponentiation error\n");
		exit(-1);
	    }
	}
	break;
    }
}

eq_table::~eq_table() {

    lt_dlclose(lib);
    lt_dlexit();
				// should free more stuff here
}

libquat_iter_func eq_table::compile() {
  
    char * file = make_code();
    char buf[1024];
    sprintf(buf, "g++ -O2 -shared -o /tmp/libquatiter.so.0.0.0 -I/usr/src/devel/libquat %s -Wl,-soname,libquatiter.so.0", file);
    system(buf);
    lib = lt_dlopen("/tmp/libquatiter.so.0.0.0");
    if(!lib) {
	fprintf(stderr, "could not compile function : %s\n", lt_dlerror());
	exit(-1);
    }	
    libquat_iter_func f = (libquat_iter_func)lt_dlsym(lib,"eq_table_iter_special");
    if(!f) {
	fprintf(stderr, "could not compile function : %s\n", lt_dlerror());
	exit(-1);
    }
    return(f);
}

				// this func spits out C code.
char * eq_table::make_code() {

    bool have_log_of_base;
    bool have_mag_of_p;

    FILE * FP;
    char * filename = strdup("/tmp/iter.cc");
    FP = fopen(filename, "w");

    fprintf(FP, "#define COMPILED\n");
    fprintf(FP, "#include <libquat.h>\n");
    fprintf(FP, "#include <libquat/equations.h>\n");
    fprintf(FP, "#define P ptr\n");

    fprintf(FP, "extern \"C\" {\n");

    if(complex_eq) {
	fprintf(FP, "#define MAG v[0]*v[0]+v[1]*v[1]\n");
    } else {
	fprintf(FP, "#define MAG v[0]*v[0]+v[1]*v[1]+v[2]*v[2]+v[3]*v[3]\n");
    }

    fprintf(FP, "int eq_table_iter_special(eq_table * ptr, double *v) {\n");

    fprintf(FP, "int iter = 0;\n");
    
    // if init needs work we need to iterate the initial table
    if(init_needs_work) {	
      // spit out variable declarations for initial table
	for(int i = init_first_work_necessary ; i < init_index_of_last_column ; i++) {
 	    fprintf(FP, "double init_value_%d_0;\n", i);
	    fprintf(FP, "double init_value_%d_1;\n", i);
	    if(!complex_eq) {
		fprintf(FP, "double init_value_%d_2;\n", i);
		fprintf(FP, "double init_value_%d_3;\n", i);
	    }
	} // last column of init is q from table
    } 

    if(c_index != -1) {
	fprintf(FP, "double iter_value_%d_0 = v[0];\n", c_index);
	fprintf(FP, "double iter_value_%d_1 = v[1];\n", c_index);
	if(!complex_eq) {
	    fprintf(FP, "double iter_value_%d_2 = v[2];\n", c_index);
	    fprintf(FP, "double iter_value_%d_3 = v[3];\n", c_index);
	}
    }
				// now iterate initial table if necessary
    if(init_needs_work) {
	have_log_of_base = false;
	have_mag_of_p = false;
	for(int i = init_first_work_necessary ; i <= init_index_of_last_column ; i++) {

#define TYPE "init"
#define TABLE init_table
#include "compile.cc"
	}
    }
    
				// output variables for table
    for(int i = first_work_necessary ; i < index_of_last_column ; i++) {
	fprintf(FP, "double iter_value_%d_0;\n", i);
	fprintf(FP, "double iter_value_%d_1;\n", i);
	if(!complex_eq) {
	    fprintf(FP, "double iter_value_%d_2;\n", i);
	    fprintf(FP, "double iter_value_%d_3;\n", i);
	}
    }
    
				// hard code initial iteration if constant
    if(!init_needs_work && init_c_index != init_index_of_last_column) {
	fprintf(FP, "v[0] = %.30f;\n", index_of_last_column, saved_init[0]);
	fprintf(FP, "v[1] = %.30f;\n", index_of_last_column, saved_init[1]);
	if(!complex_eq) {
	    fprintf(FP, "v[2] = %.30f;\n", 
		    index_of_last_column, saved_init[2]);
	    fprintf(FP, "v[3] = %.30f;\n", 
		    index_of_last_column, saved_init[3]);
	}
    }

    fprintf(FP, "do {\n");
    if(debug) {
	fprintf(FP, "printf(\"iter %%d: %%.2f %%.2f\", iter, v[0], v[1]);\n");
	if(!complex_eq)
	    fprintf(FP, "printf(\"%%.2f %%.2f\", v[2], v[3]);\n");
	fprintf(FP, "printf(\"\\n\");\n");
    }


    have_log_of_base = false;
    have_mag_of_p = false;
    
    for(int i = first_work_necessary ; i <= index_of_last_column ; i++) {

#define TABLE table
#define TYPE "iter"
#include "compile.cc"
    }

    fprintf(FP, "} while(++iter < ptr->max_iter && ");

    if(btest_greater)
	fprintf(FP, "MAG > ptr->max_val);\n", index_of_last_column);
    else
	fprintf(FP, "MAG < ptr->max_val);\n", index_of_last_column);

    fprintf(FP, "return(iter);}}");
    fclose(FP);
    return(filename);
}

libquat_iter_func_1 eq_table::compile_1() {
  
    char * file = make_code_1();
    char buf[1024];
    sprintf(buf, "g++ -O2 -shared -o /tmp/libquatiter_1.so.0.0.0 -I/usr/src/devel/libquat %s -Wl,-soname,libquatiter_1.so.0", file);
    system(buf);
    lib = lt_dlopen("/tmp/libquatiter_1.so.0.0.0");
    if(!lib) {
	fprintf(stderr, "could not compile function : %s\n", lt_dlerror());
	exit(-1);
    }	
    libquat_iter_func_1 f = (libquat_iter_func_1)lt_dlsym(lib,"eq_table_iter_special_1");
    if(!f) {
	fprintf(stderr, "could not compile function : %s\n", lt_dlerror());
	exit(-1);
    }
    return(f);
}
				// this func spits out C code.
				// unfortunately it is not reentrant (statics)
char * eq_table::make_code_1() {

    bool have_log_of_base;
    bool have_mag_of_p;

    FILE * FP;
    char * filename = strdup("/tmp/iter_1.cc");
    FP = fopen(filename, "w");

    fprintf(FP, "#define COMPILED\n");
    fprintf(FP, "#include <libquat.h>\n");
    fprintf(FP, "#include <libquat/equations.h>\n");
    fprintf(FP, "#define P ptr\n");

    fprintf(FP, "extern \"C\" {\n");

    fprintf(FP, "void eq_table_iter_special_1(double *v, bool init) {\n");

    fprintf(FP, "int iter = 0;\n");
    
    // if init needs work we need to iterate the initial table
    if(init_needs_work) {	
      // spit out variable declarations for initial table
	for(int i = init_first_work_necessary ; i < init_index_of_last_column ; i++) {
 	    fprintf(FP, "double init_value_%d_0;\n", i);
	    fprintf(FP, "double init_value_%d_1;\n", i);
	    if(!complex_eq) {
		fprintf(FP, "double init_value_%d_2;\n", i);
		fprintf(FP, "double init_value_%d_3;\n", i);
	    }
	} // last column of init is q from table
    } 

    if(c_index != -1) {
	fprintf(FP, "static double iter_value_%d_0;\n", c_index);
	fprintf(FP, "static double iter_value_%d_1;\n", c_index);
	if(!complex_eq) {
	    fprintf(FP, "static double iter_value_%d_2;\n", c_index);
	    fprintf(FP, "static double iter_value_%d_3;\n", c_index);
	}
    }

    fprintf(FP, "if(init) {\n");

    if(c_index != -1) {
	fprintf(FP, "iter_value_%d_0 = v[0];\n", c_index);
	fprintf(FP, "iter_value_%d_1 = v[1];\n", c_index);
	if(!complex_eq) {
	    fprintf(FP, "iter_value_%d_2 = v[2];\n", c_index);
	    fprintf(FP, "iter_value_%d_3 = v[3];\n", c_index);
	}
    }

				// now iterate initial table if necessary
    if(init_needs_work) {
	have_log_of_base = false;
	have_mag_of_p = false;
	for(int i = init_first_work_necessary ; i <= init_index_of_last_column ; i++) {

#define TYPE "init"
#define TABLE init_table
#include "compile.cc"
	}
    }

				// hard code initial iteration if constant
    if(!init_needs_work && init_c_index != init_index_of_last_column) {
	fprintf(FP, "v[0] = %.30f;\n", index_of_last_column, saved_init[0]);
	fprintf(FP, "v[1] = %.30f;\n", index_of_last_column, saved_init[1]);
	if(!complex_eq) {
	    fprintf(FP, "v[2] = %.30f;\n", 
		    index_of_last_column, saved_init[2]);
	    fprintf(FP, "v[3] = %.30f;\n", 
		    index_of_last_column, saved_init[3]);
	}
    }

    
    fprintf(FP, "} else {\n");

				// output variables for table
    for(int i = first_work_necessary ; i < index_of_last_column ; i++) {
	fprintf(FP, "double iter_value_%d_0;\n", i);
	fprintf(FP, "double iter_value_%d_1;\n", i);
	if(!complex_eq) {
	    fprintf(FP, "double iter_value_%d_2;\n", i);
	    fprintf(FP, "double iter_value_%d_3;\n", i);
	}
    }
    

//    fprintf(FP, "do {\n");
    if(debug) {
	fprintf(FP, "printf(\"iter %%d: %%.2f %%.2f\", iter, v[0], v[1]);\n");
	if(!complex_eq)
	    fprintf(FP, "printf(\"%%.2f %%.2f\", v[2], v[3]);\n");
	fprintf(FP, "printf(\"\\n\");\n");
    }


    have_log_of_base = false;
    have_mag_of_p = false;
    
    for(int i = first_work_necessary ; i <= index_of_last_column ; i++) {

#define TABLE table
#define TYPE "iter"
#include "compile.cc"
    }

//    fprintf(FP, "} while(++iter < ptr->max_iter && ");

//     if(btest_greater)
// 	fprintf(FP, "MAG > ptr->max_val);\n", index_of_last_column);
//     else
// 	fprintf(FP, "MAG < ptr->max_val);\n", index_of_last_column);

//    fprintf(FP, "return(iter);}}");
    fprintf(FP, "}}}");
    fclose(FP);
    return(filename);
}

void eq_table::schedule(expr_ref expr, int rank, Entry ** list) {

    if(expr->rank == rank)
	add_expr(expr,list);

    if(expr->rchld)
	schedule(expr->rchld, rank, list);
    if(expr->lchld)
	schedule(expr->lchld, rank, list);
}

expr_ref eq_table::parse_expr(char * expr) {
    yy_scan_string(expr);
    if(yyparse()==0)
	return(tree_top);
    else
	return NULL;
}

eq_table::eq_table(const char * expr_string, const char * init_expr_string,
		   int _max_iter, double _max_val) {

    if(!init_expr_string) 
	init_expr_string = "0";

    btest_greater = false;	// while (mag < bailout) iterate
  
    expr_ref expr = parse_expr(const_cast<char *>(expr_string));
    expr_ref init_expr = parse_expr(const_cast<char *>(init_expr_string));

    if(strchr(expr_string, 'z'))
	complex_eq = true;
    else
	complex_eq = false;

//    printf("complex = %d\n", complex_eq);

    if(complex_eq) {
//	printf("forcing complex\n");
	expr->force_complex();
	init_expr->force_complex();
    }

    init_expr->assign_retvals(complex_eq);

    max_iter = _max_iter;
    max_val = _max_val;

    max_val *= max_val;		// we compare against maxval^2 to avoid sqrt

    debug = false;
    
    table_list = 0;
    init_c_index = -1;
    init_table_list = 0;
    c_index = -1;
    q_index = -1;
    num_iterates = 1;

    //    expr->dump();

				// initial table (for q_0)
    for(int i = 0 ; i <= init_expr->rank ; i++) 
	schedule(init_expr, i, &init_table_list);

    Entry * old_table;

    for(old_table = init_table_list ; old_table && old_table->next;
	old_table = old_table->next);
    
    init_index_of_last_column = old_table->index;
    old_table = init_table_list;
    init_table = new Entry *[init_index_of_last_column+1];
    for(int i = 0 ; i <= init_index_of_last_column ; i++) {
	init_table[i] = old_table;
	old_table = old_table->next;
    }

    deal_with_constants(init_table, init_index_of_last_column);

    init_first_work_necessary = init_index_of_last_column + 1;
    for(int i = 0 ; i <= init_index_of_last_column ; i++)
	if(init_table[i]->expr->rank) {
	    init_first_work_necessary = i;
	    break;
	}

				// main table
    for(int i = 0 ; i <= expr->rank ; i++)
	schedule(expr, i, &table_list);

    for(old_table = table_list ; old_table && old_table->next;
	old_table = old_table->next);

    index_of_last_column = old_table->index;
    old_table = table_list;

    table = new Entry *[index_of_last_column+1];
    for(int i = 0 ; i <= index_of_last_column ; i++) {
	table[i] = old_table;
	old_table = old_table->next;
    }

    deal_with_constants(table, index_of_last_column);
    deal_with_old_iterates();

    first_work_necessary = index_of_last_column + 1;
    for(int i = 0 ; i <= index_of_last_column ; i++)
	if(table[i]->expr->rank) {
	    first_work_necessary = i;
	    break;
	}

    if(init_table[init_index_of_last_column]->expr->rank == 0 &&
       init_c_index != init_index_of_last_column) {
	init_needs_work = false;
#if 0
	memcpy(saved_init, init_table[init_index_of_last_column]->vect, 
	       sizeof(double)*4);
#else
 	saved_init[0] = init_table[init_index_of_last_column]->vect[0];
 	saved_init[1] = init_table[init_index_of_last_column]->vect[1];
 	saved_init[2] = init_table[init_index_of_last_column]->vect[2];
 	saved_init[3] = init_table[init_index_of_last_column]->vect[3];
#endif
    } else {
	old_iterates[0]->index = init_table[init_index_of_last_column]->index;
	old_iterates[0]->expr = init_table[init_index_of_last_column]->expr;
	old_iterates[0]->rule = init_table[init_index_of_last_column]->rule;
	old_iterates[0]->dep1 = init_table[init_index_of_last_column]->dep1;
	old_iterates[0]->dep2 = init_table[init_index_of_last_column]->dep2;
	
	init_table[init_index_of_last_column] = old_iterates[0]; // doh!!
	init_needs_work = true;
    }
    
				// output
#if 0
    for(int i = 0 ; i <= index_of_last_column ; i++)
	printf("%d\t", table[i]->index);
    printf("\n");
    for(int i = 0 ; i <= index_of_last_column ; i++)
	printf("%s\t", table[i]->expr->data);
    printf("\n");
    for(int i = 0 ; i <= index_of_last_column ; i++)
	printf("%d\t", table[i]->rule);
    printf("\n");
    for(int i = 0 ; i <= index_of_last_column ; i++)
	printf("%d\t", table[i]->dep1);
    printf("\n");
    for(int i = 0 ; i <= index_of_last_column ; i++)
	printf("%d\t", table[i]->dep2);
    printf("\n");
#endif
} 

void eq_table::deal_with_constants(Entry ** tab, int max) {

				// look through table and insert
				// values for all the literals
    for(int i = 0 ; i <= max ; i++) {
	if(tab[i]->expr->token_code == TOK_LIT_QUAT ||
	   tab[i]->expr->token_code == TOK_LIT_COMPLEX ||
	   tab[i]->expr->token_code == 'i' ||
	   tab[i]->expr->token_code == 'j' ||
	   tab[i]->expr->token_code == 'k' ||
	   tab[i]->expr->token_code == TOK_LIT_SCALAR ||
	   tab[i]->expr->token_code == TOK_LIT_PI ||
	   tab[i]->expr->token_code == 'e') {

	    tab[i]->vect[0] = tab[i]->expr->val->s;
	    tab[i]->vect[1] = tab[i]->expr->val->a;
	    tab[i]->vect[2] = tab[i]->expr->val->b;

	    tab[i]->vect[3] = tab[i]->expr->val->c;
	}
    }    
}


void eq_table::deal_with_old_iterates() {
				
    old_iterates = new Entry*[num_iterates];

    for(int i = 0 ; i < num_iterates ; i++)
	old_iterates[i] = NULL;

    if(q_index != -1) {
	old_iterates[0] = table[q_index];
    }
    
    expr_ref e;
    
    for(int i = 0 ; i <= index_of_last_column ; i++)
	if(table[i]->expr->token_code == '_') {
	    e = table[i]->expr;
	    while(e->rchld) e = e->rchld;
	    old_iterates[atoi(e->data)] = table[i];
	    //	    printf("found iterate %d\n", atoi(e->data));
	}
    
    for(int i = 0 ; i < num_iterates ; i++)
	if(old_iterates[i] == NULL)
	    old_iterates[i] = new Entry();
}

int eq_table::get_column(expr_ref expr, Entry ** list) {

    Entry * ptr;
				// look for expr in table
    if(*list)			// should not be necessary
	for(ptr = *list ; ptr ; ptr = ptr->next) 
	    if(ptr->expr->eql(expr))
		return ptr->index;

    return(-1);
}

eq_table::Entry * eq_table::add_column(expr_ref expr, Entry ** list) {

    int index = get_column(expr, list);

    if(index == -1) {
	Entry * new_entry = new Entry();

	new_entry->expr = expr;
	Entry * ptr;

	if(*list) {
	    for(ptr = *list ; ptr->next ; ptr = ptr->next);
	    new_entry->index = ptr->index + 1;
	    new_entry->prev = ptr;
	    ptr->next = new_entry;
	} else {
	    new_entry->index = 0;
	    *list = new_entry;
	}
//	printf("entry is %p\n", new_entry);

	return new_entry;
    } else {
	Entry * ans;
	for(Entry * ptr = *list ; ptr ; ptr = ptr->next)
	    if(index == ptr->index) {
		ans = ptr;
		break;
	    }

	return(ans);
    }
}

inline double eq_table::mag() {	// returns mag^2
    return(table[index_of_last_column]->vect[0]*
	   table[index_of_last_column]->vect[0]+
	   table[index_of_last_column]->vect[1]*
	   table[index_of_last_column]->vect[1]+
	   table[index_of_last_column]->vect[2]*
	   table[index_of_last_column]->vect[2]+
	   table[index_of_last_column]->vect[3]*
	   table[index_of_last_column]->vect[3]);
}

#define SIGNATURE int eq_table::iter(double *v) 
#define RETVALS return(iter);
#define P this
#include "eval.cc"

#define SIGNATURE int eq_table::iter_debug(double *v) 
#ifndef OUTPUT
#define OUTPUT
#define REMOVE_OUTPUT
#endif
#define RETVALS return(iter);
#define P this
#include "eval.cc"
#ifdef REMOVE_OUTPUT
#undef OUTPUT
#endif

#define SOUND

#define P this
double * eq_table::iter_init(double * v) {

    eq_table::Entry * curr;

    if(P->init_c_index != -1) {
	P->init_table[P->init_c_index]->vect[0] = v[0];
	P->init_table[P->init_c_index]->vect[1] = v[1];
	P->init_table[P->init_c_index]->vect[2] = v[2];
	P->init_table[P->init_c_index]->vect[3] = v[3];
    }

    if(P->init_needs_work) {

#define TABLE init_table
#define INIT init_first_work_necessary
#define END init_index_of_last_column
#include "iter.cc"

    } else if(P->init_c_index != P->init_index_of_last_column) {
#if 0
	memcpy(P->old_iterates[0]->vect, P->saved_init, sizeof(double)*4);
#else
	P->old_iterates[0]->vect[0] = P->saved_init[0];
	P->old_iterates[0]->vect[1] = P->saved_init[1];
	P->old_iterates[0]->vect[2] = P->saved_init[2];
	P->old_iterates[0]->vect[3] = P->saved_init[3];
#endif
    }

    for(int i = 1 ; i < P->num_iterates ; i++)
	memset(P->old_iterates[i]->vect, 0, sizeof(double)*4);    

    // try to combine this with one above at some point
    if(P->c_index != -1) {
      P->table[P->c_index]->vect[0] = v[0];
	P->table[P->c_index]->vect[1] = v[1];
	P->table[P->c_index]->vect[2] = v[2];
	P->table[P->c_index]->vect[3] = v[3];
    }

    return P->old_iterates[0]->vect;
}

double * eq_table::iter_1() {

    eq_table::Entry * curr;


#define TABLE table
#define INIT first_work_necessary
#define END index_of_last_column
#include "iter.cc"

    for(int i =  P->num_iterates - 1 ; i > 0 ; i--) {
	P->old_iterates[i]->vect[0] = P->old_iterates[i-1]->vect[0];
	P->old_iterates[i]->vect[1] = P->old_iterates[i-1]->vect[1];
	P->old_iterates[i]->vect[2] = P->old_iterates[i-1]->vect[2];
	P->old_iterates[i]->vect[3] = P->old_iterates[i-1]->vect[3];
    }
	
    P->old_iterates[0]->vect[0] = P->table[P->index_of_last_column]->vect[0];
    P->old_iterates[0]->vect[1] = P->table[P->index_of_last_column]->vect[1];
    P->old_iterates[0]->vect[2] = P->table[P->index_of_last_column]->vect[2];
    P->old_iterates[0]->vect[3] = P->table[P->index_of_last_column]->vect[3];

    return(P->old_iterates[0]->vect);
}
#undef P
#undef SOUND
				// c interface follows:
extern "C" {

    eq_table * eq_table_new(char * s, char * i) { 
	return(new eq_table(s, i, 400, 40));
    }

#define SIGNATURE int eq_table_iter(eq_table * ptr, double * v)
#define RETVALS v[0] = ptr->table[ptr->index_of_last_column]->vect[0]; \
v[1] = ptr->table[ptr->index_of_last_column]->vect[0]; \
v[2] = ptr->table[ptr->index_of_last_column]->vect[0]; \
v[3] = ptr->table[ptr->index_of_last_column]->vect[0]; \
	return iter;
#define P ptr
#include "eval.cc"
    
    libquat_iter_func libquat_compile(eq_table * p) {
	return(p->compile());
    }
}
