#ifndef EQ_TABLE_H
#define EQ_TABLE_H

#include <libquat.h>
#include <libquat/expr_tree.h>
#include <ltdl.h>
#define LIBQUAT_GT 1
#define LIBQUAT_LT 0

class eq_table {

private:
public:

    class Entry {
    public:
	int rule, dep1, dep2;	    

	double vect[4];		// actual values
	int index;
	Entry * next, * prev;
	expr_ref expr;
	bool skip;
	void swap(Entry ** front);

	Entry(): next(0), prev(0) {
	    vect[0] = vect[1] = vect[2] = vect[3] = 0;
	}
	Entry(Entry*);
    };
    
    int max_iter;
    double max_val;
    int num_iterates;		// number of iterates to keep around
    Entry ** old_iterates;
    
    lt_dlhandle lib;
    Entry * table_list;
    Entry ** table;
    int index_of_last_column;
    int first_work_necessary;
    int c_index;		// index of c (position specific constant)
    int q_index;		// index of q 

    Entry * init_table_list;
    Entry ** init_table;
    int init_index_of_last_column;
    int init_first_work_necessary;
    int init_c_index;
    
    bool debug;			
    bool complex_eq;		// true if var is z
    bool btest_greater;		// 1 tests with mag > bailout
				// 0 tests with mag < bailout
    bool init_needs_work;	// false if top of init tree is schedule 0
    double saved_init[4];

    void add_expr(expr_ref expr, Entry ** list);
    void schedule(expr_ref expr, int rank, Entry ** list);
    Entry * add_column(expr_ref expr, Entry ** list);
    int get_column(expr_ref expr, Entry ** list);
    void deal_with_old_iterates();
    void deal_with_constants(Entry ** tab, int max);

    expr_ref parse_expr(char *);
    eq_table(eq_table*);
    ~eq_table();
    eq_table(const char * expr, const char * init_expr, int max_iter, double max_val);
    int iter(double *);
    int iter_debug(double *);
    double * iter_init(double *);
    double * iter_1();
    libquat_iter_func compile();
    char * make_code();
    libquat_iter_func_1 compile_1();
    char * make_code_1();
    double mag();
    void set_debug() { debug = true; }
 /*     void set_init_q_val(double * v); */
/*     void set_q_to_c_initially(); */
    void set_btest(int which) {btest_greater = which;}
};

#endif
