#include "eq_table.h"
#include "expr_parser.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define Q_PLUS_Q 0
#define Q_PLUS_S 1
#define S_PLUS_S 2

#define Q_MINUS_Q 3
#define Q_MINUS_S 4
#define S_MINUS_S 5

#define Q_TIMES_Q 6
#define Q_TIMES_S 7
#define S_TIMES_S 8

#define Q_DIV_S	 9
#define S_DIV_S	 10

#define Q_EXP_S_EVEN 11
#define S_EXP_Q	12
#define S_EXP_S 17

#define RE_Q    13
#define IM_I_Q	14
#define IM_J_Q	15
#define IM_K_Q	16

#define ABS     18

eq_table::eq_table(eq_table *copy) {

    max_iter = copy->max_iter;
    max_val = copy->max_val;
    num_iterates = copy->num_iterates;
    index_of_last_column = copy->index_of_last_column;
    c_index = copy->c_index;
    q_index = copy->q_index;
    //q_index = copy->q_index;
    table = new Entry*[index_of_last_column+1];
    for(int i = 0 ; i <= index_of_last_column ; i++) {
	table[i] = new Entry(copy->table[i]);
    }
    deal_with_old_iterates();
}

eq_table::Entry::Entry(Entry * copy) {
    
    for(int i = 0 ; i < 4 ; i++) {
	vect[i] = copy->vect[i];
    }
    index = copy->index;
    expr = copy->expr;
    skip = copy->skip;
}

void eq_table::schedule(expr_ref expr, int rank) {

    if(expr->rank == rank) {
	

    }

    if(expr->rchld)
	schedule(expr->rchld, rank);
    if(expr->lchld)
	schedule(expr->lchld, rank);
}

eq_table::eq_table(expr_ref expr, int _max_iter, double _max_val) {
  
    max_iter = _max_iter;
    max_val = _max_val;

    max_val *= max_val;
    
    table_list = 0;

    c_index = -1;
    q_index = -1;
    num_iterates = 1;
    expr_ref e;
    
    printf("expr is ");
    expr->dump();
    printf("\n");

    int val;

    add_column(expr);
    
    for(Entry * ptr = table_list ; ptr ; ptr = ptr->next) {
	switch (ptr->expr->token_code) {

	case TOK_RE:
	    printf("doing re\n");
	    ptr->rule = RE_Q;
	    ptr->dep1 = add_column(ptr->expr->rchld);
	    ptr->dep2 = -1;
	    break;

	case TOK_IM_I:
	    printf("doing im\n");
	    ptr->rule = IM_I_Q;
	    ptr->dep1 = add_column(ptr->expr->rchld);
	    ptr->dep2 = -1;
	    break;

	case TOK_IM_J:
	    ptr->rule = IM_J_Q;
	    ptr->dep1 = add_column(ptr->expr->rchld);
	    ptr->dep2 = -1;
	    break;

	case TOK_IM_K:
	    ptr->rule = IM_K_Q;
	    ptr->dep1 = add_column(ptr->expr->rchld);
	    ptr->dep2 = -1;
	    break;

	case '_':		// subscript
	    e = ptr->expr;
	    while(e->rchld) e = e->rchld;
	    
	    printf("token is %d\n", e->token_code);
	    printf("value is %d\n", e->val->s);
	    printf("data is %s\n", e->data);
	    val = atoi(e->data);
	    if(!e->token_code == TOK_LIT_SCALAR) {
		fprintf(stderr, "scalar not found in subscript\n");
		exit(-1);
	    }
	    if(num_iterates < val + 1)
		num_iterates = val + 1;
	    break;

	case 'c':
	    c_index = ptr->index;
	case 'q':
	    if(ptr->expr->token_code == 'q')
		q_index = ptr->index;
	case 'i':
	case 'j':
	case 'k':
	case TOK_LIT_QUAT:
	    break;
		
	case 'e':
	case TOK_LIT_PI:
	case TOK_LIT_SCALAR:
	    break;

	case '+':
	    if(ptr->expr->lchld->scalar) 
		if(ptr->expr->rchld->scalar) {
		    printf("doing s+s\n");
		    ptr->rule = S_PLUS_S;
		    ptr->dep1 = add_column(ptr->expr->lchld);
		    ptr->dep2 = add_column(ptr->expr->rchld);
		    printf("dep1 = %s dep2 = %s\n", ptr->expr->lchld->data, 
			   ptr->expr->rchld->data);
		} else {
		    printf("doing s+q\n");
		    ptr->rule = Q_PLUS_S;
		    ptr->dep1 = add_column(ptr->expr->rchld);
		    ptr->dep2 = add_column(ptr->expr->lchld);
		    printf("dep1 = %s dep2 = %s\n", ptr->expr->lchld->data, 
			   ptr->expr->rchld->data);

		}
	    else 
		if(ptr->expr->rchld->scalar) {
		    printf("doing q+s\n");
		    ptr->rule = Q_PLUS_S;
		    ptr->dep1 = add_column(ptr->expr->lchld);
		    ptr->dep2 = add_column(ptr->expr->rchld);
		} else {
		    printf("doing q+q\n");
		    ptr->rule = Q_PLUS_Q;
		    ptr->dep1 = add_column(ptr->expr->lchld);
		    ptr->dep2 = add_column(ptr->expr->rchld);
		}
	    break;

	case '-':
	    if(ptr->expr->lchld->scalar) 
		if(ptr->expr->rchld->scalar) {
		    printf("doing s-s\n");
		    ptr->rule = S_MINUS_S;
		    ptr->dep1 = add_column(ptr->expr->lchld);
		    ptr->dep2 = add_column(ptr->expr->rchld);
		} else {
		    printf("doing s-q\n");
		    ptr->rule = Q_MINUS_S;
		    ptr->dep1 = add_column(ptr->expr->rchld);
		    ptr->dep2 = add_column(ptr->expr->lchld);
		    printf("dep1 = %s dep2 = %s\n", ptr->expr->lchld->data, 
			   ptr->expr->rchld->data);

		}
	    else 
		if(ptr->expr->rchld->scalar) {
		    printf("doing q-s\n");
		    printf("dep1 = %s dep2 = %s\n", ptr->expr->lchld->data, 
			   ptr->expr->rchld->data);
		    ptr->rule = Q_MINUS_S;
		    ptr->dep1 = add_column(ptr->expr->lchld);
		    ptr->dep2 = add_column(ptr->expr->rchld);
		} else {
		    printf("doing q-q\n");
		    ptr->rule = Q_MINUS_Q;
		    ptr->dep1 = add_column(ptr->expr->lchld);
		    ptr->dep2 = add_column(ptr->expr->rchld);
		    printf("dep1 = %s dep2 = %s\n", ptr->expr->lchld->data, 
			   ptr->expr->rchld->data);

		}
	    break;
				
	case '*':
	    if(ptr->expr->lchld->scalar) {
		if(ptr->expr->rchld->scalar) { // scalar*scalar
		    printf("doing s * s");
		    ptr->rule = S_TIMES_S;
		    ptr->dep1 = add_column(ptr->expr->lchld);
		    ptr->dep2 = add_column(ptr->expr->rchld);
		    printf(" (%d %s %d %s)\n", ptr->dep1, ptr->expr->lchld->data, ptr->dep2, ptr->expr->rchld->data);
		} else {	// scalar*quat
		    printf("doing s * q\n");
		    printf("dep1 = %s dep2 = %s\n", ptr->expr->rchld->data, ptr->expr->lchld->data);
		    ptr->rule = Q_TIMES_S;
		    ptr->dep1 = add_column(ptr->expr->rchld);
		    ptr->dep2 = add_column(ptr->expr->lchld);
		}
	    } else {
		if(ptr->expr->rchld->scalar) { // quat*scalar
		    printf("doing q * s\n");
		    printf("dep1 = %s dep2 = %s\n", ptr->expr->lchld->data, ptr->expr->rchld->data);
		    ptr->rule = Q_TIMES_S;
		    ptr->dep1 = add_column(ptr->expr->lchld);
		    ptr->dep2 = add_column(ptr->expr->rchld);
		} else {	// quaternion mult
		    printf("doing q * q\n");
		    ptr->rule = Q_TIMES_Q;
		    ptr->dep1 = add_column(ptr->expr->lchld);
		    ptr->dep2 = add_column(ptr->expr->rchld);
		}
	    }
	    break;

	case '/':
	    if(!ptr->expr->lchld->scalar) {
		printf("right side of division must be scalar!\n");
		exit(-1);
	    }
	    if(ptr->expr->rchld->scalar) {
		ptr->rule = S_DIV_S;
		ptr->dep1 = add_column(ptr->expr->lchld);
		ptr->dep2 = add_column(ptr->expr->rchld);
	    } else {
		ptr->rule = Q_DIV_S;
		ptr->dep1 = add_column(ptr->expr->lchld);
		ptr->dep2 = add_column(ptr->expr->rchld);
	    }
	    break;

	case '^':
	    if(ptr->expr->lchld->scalar) { // scalar base
		if(ptr->expr->rchld->scalar)
		    ptr->rule = S_EXP_S;
		else
		    ptr->rule = S_EXP_Q;
		ptr->dep1 = add_column(ptr->expr->lchld);
		ptr->dep2 = add_column(ptr->expr->rchld);
	    } else {		// quaternion base
		if(((int)ptr->expr->rchld->val->s) % 2) { // odd exponent
		    ptr->rule = Q_TIMES_Q;
		    ptr->dep1 = add_column(ptr->expr->lchld);

		    char buf[30];

		    expr_ref new_child = ptr->expr->copy();
		    sprintf(buf, "%d", int(ptr->expr->rchld->val->s)-1);
		    new_child->rchld->data = strdup(buf);
		    new_child->rchld->val->s = ptr->expr->rchld->val->s - 1;
		    
		    ptr->dep2 = add_column(new_child);

		} else {		// even exponent
		    ptr->rule = Q_EXP_S_EVEN;
					    
		    char buf[30];
		    expr_ref new_child = ptr->expr->copy();
		    int power = int(new_child->rchld->val->s)/2;
		    if(power == 1) {
			ptr->dep1 = add_column(new_child->lchld);
			delete new_child->rchld;
			delete new_child;
		    } else {
			sprintf(buf, "%d", power);
			new_child->rchld->data = strdup(buf);
			new_child->rchld->val->s = power;
			ptr->dep1 = add_column(new_child);
		    }
		}
	    }
	    break;
	}
    }

    Entry * old_table = table_list;

    for(old_table=table_list;old_table->next;old_table=old_table->next);

    index_of_last_column = old_table->index;
    old_table = table_list;

    table = new Entry *[index_of_last_column+1];
    for(int i = 0 ; i <= index_of_last_column ; i++) {
	table[i] = old_table;
	old_table = old_table->next;
    }

    printf("table = %p\n", table);

				// now we need to look through table and insert
				// values for all the literals
    for(int i = 0 ; i <= index_of_last_column ; i++) {
	if(table[i]->expr->token_code == TOK_LIT_QUAT ||
	   table[i]->expr->token_code == 'i' ||
	   table[i]->expr->token_code == 'j' ||
	   table[i]->expr->token_code == 'k' ||
	   table[i]->expr->token_code == TOK_LIT_SCALAR ||
	   table[i]->expr->token_code == TOK_LIT_PI ||
	   table[i]->expr->token_code == 'e') {

	    table[i]->skip = true;
	    table[i]->vect[0] = table[i]->expr->val->s;
	    table[i]->vect[1] = table[i]->expr->val->a;
	    table[i]->vect[2] = table[i]->expr->val->b;
	    table[i]->vect[3] = table[i]->expr->val->c;
	} else {
	    table[i]->skip = false;
	}
    }    

    deal_with_old_iterates();

    if(c_index != -1)
	table[c_index]->skip = true;    
} 

void eq_table::deal_with_old_iterates() {
				
    old_iterates = new Entry*[num_iterates];

    for(int i = 0 ; i < num_iterates ; i++)
	old_iterates[i] = NULL;

    printf("%d iterates\n", num_iterates);
    
    if(q_index != -1) {
	old_iterates[0] = table[q_index];
	table[q_index]->skip = true;
    }
    
    expr_ref e;
    
    for(int i = 0 ; i <= index_of_last_column ; i++)
	if(table[i]->expr->token_code == '_') {
	    e = table[i]->expr;
	    while(e->rchld) e = e->rchld;
	    printf("found itate %.2f\n", e->val->s);
	    old_iterates[atoi(e->data)] = table[i];
	    table[i]->skip = true;
	}
    
    for(int i = 0 ; i < num_iterates ; i++)
	if(old_iterates[i] == NULL)
	    old_iterates[i] = new Entry();
}

int eq_table::add_column(expr_ref expr) {

    Entry * ptr;
				// look for expr in table
    for(ptr = table_list ; ptr ; ptr = ptr->next) 
	if(ptr->expr->eql(expr))
	    return ptr->index;

				// expr not found in table 
    Entry * new_entry = new Entry();

    new_entry->expr = expr;

    if(table_list) {
	for(ptr = table_list ; ptr->next ; ptr = ptr->next);
    	new_entry->index = ptr->index + 1;
	ptr->next = new_entry;
    } else {
	new_entry->index = 0;
	table_list = new_entry;
    }

    return new_entry->index;
}

double eq_table::iter(double s, double a, double b, double c, int & iters) {

  //    printf("iter-- q_index = %d this = %p\n", q_index, this);
    
#if 1
    old_iterates[0]->vect[0] = s; // should detect if mandelbrot and use 0 
    old_iterates[0]->vect[1] = a; // instead.
    old_iterates[0]->vect[2] = b;
    old_iterates[0]->vect[3] = c;
#else
    old_iterates[0]->vect[0] = 0; // should detect if mandelbrot and use 0 
    old_iterates[0]->vect[1] = 0; // instead.
    old_iterates[0]->vect[2] = 0;
    old_iterates[0]->vect[3] = 0;
#endif

    for(int i = 1 ; i < num_iterates ; i++)
	memset(old_iterates[i]->vect, 0, sizeof(double)*4);    

    if(c_index != -1) {
	table[c_index]->vect[0] = s;
	table[c_index]->vect[1] = a;
	table[c_index]->vect[2] = b;
	table[c_index]->vect[3] = c;
    }

//     printf("p is %.2f %.2f %.2f %.2f\n", s, a, b, c);

    int iter = 0;

    do {
	
#ifdef OUTPUT
	printf("%d: %.2f %.2f %.2f %.2f\n", iter, old_iterates[0]->vect[0],
	       old_iterates[0]->vect[1], old_iterates[0]->vect[2], 
	       old_iterates[0]->vect[3]);
#endif
	
	eval();
	
#if 0
	for(int i = num_iterates - 1 ; i > 0 ; i--) {
	    memcpy(old_iterates[i-1],
		   old_iterates[i],
		   sizeof(double)*4);
	}

	memcpy(table[0]->vect,
	       old_iterates[0]->vect,
	       sizeof(double)*4);

#else

#if 0
	for(int i = 0 ; i < num_iterates ; i++)
	    printf("%.2f %.2f %.2f %.2f\n", old_iterates[i]->vect[0], 
		   old_iterates[i]->vect[1], old_iterates[i]->vect[2],
		   old_iterates[i]->vect[3]); 
	printf("\n");
#endif
	for(int i = iter > num_iterates - 1 ? num_iterates - 1 : iter ;
	    i > 0 ; i--) {
	    old_iterates[i]->vect[0] = old_iterates[i-1]->vect[0];
	    old_iterates[i]->vect[1] = old_iterates[i-1]->vect[1];
	    old_iterates[i]->vect[2] = old_iterates[i-1]->vect[2];
	    old_iterates[i]->vect[3] = old_iterates[i-1]->vect[3];
	}
	
	old_iterates[0]->vect[0] = table[0]->vect[0];
	old_iterates[0]->vect[1] = table[0]->vect[1];
	old_iterates[0]->vect[2] = table[0]->vect[2];
	old_iterates[0]->vect[3] = table[0]->vect[3];
#endif

    } while(++iter < max_iter && mag() < max_val);
    
#ifdef OUTPUT
    printf("%d: %.2f %.2f %.2f %.2f\n", iter, old_iterates[0]->vect[0],
	   old_iterates[0]->vect[1], old_iterates[0]->vect[2], 
	   old_iterates[0]->vect[3]);
#endif

    iters = iter;		// pass out iterations

    return(mag());
    //    printf("iter return\n");
}

inline double eq_table::mag() {	// returns mag^2
    return(table[0]->vect[0]*table[0]->vect[0]+
	   table[0]->vect[1]*table[0]->vect[1]+
	   table[0]->vect[2]*table[0]->vect[2]+
	   table[0]->vect[3]*table[0]->vect[3]);
}

void eq_table::eval() {

    Entry * curr;

    for(int i = index_of_last_column ; i >= 0 ; i--) {
	if(table[i]->skip)
	    continue;
	curr = table[i];

	    int larg = curr->dep1;
	    int rarg = curr->dep2;
	    double log_of_base;
	    
	    switch (curr->rule) {

	    case Q_PLUS_Q:	// (x+y) (quaternion)
		curr->vect[0] = table[larg]->vect[0] + table[rarg]->vect[0];
		curr->vect[1] = table[larg]->vect[1] + table[rarg]->vect[1];
		curr->vect[2] = table[larg]->vect[2] + table[rarg]->vect[2];
		curr->vect[3] = table[larg]->vect[3] + table[rarg]->vect[3];
		break;

	    case Q_PLUS_S:	// x+y (scalar)
		curr->vect[0] = table[larg]->vect[0] + table[rarg]->vect[0];
		curr->vect[1] = table[larg]->vect[1];
		curr->vect[2] = table[larg]->vect[2];
		curr->vect[3] = table[larg]->vect[3];
		break;
		
	    case S_PLUS_S:	// x+y scalar
		curr->vect[0] = table[larg]->vect[0] + table[rarg]->vect[0];
		break;

	    case Q_MINUS_Q:	// (x-y) (quaternion)
		curr->vect[0] = table[larg]->vect[0] - table[rarg]->vect[0];
		curr->vect[1] = table[larg]->vect[1] - table[rarg]->vect[1];
		curr->vect[2] = table[larg]->vect[2] - table[rarg]->vect[2];
		curr->vect[3] = table[larg]->vect[3] - table[rarg]->vect[3];
		break;

	    case Q_MINUS_S:	// x-y (scalar)
		curr->vect[0] = table[larg]->vect[0] - table[rarg]->vect[0];
		curr->vect[1] = table[larg]->vect[1];
		curr->vect[2] = table[larg]->vect[2];
		curr->vect[3] = table[larg]->vect[3];
		break;

	    case S_MINUS_S:	// x-y (scalar)
		curr->vect[0] = table[larg]->vect[0] - table[rarg]->vect[0];
		break;

	    case Q_TIMES_Q:	// (x*y) (quaternion)
		curr->vect[0] = 
		    table[larg]->vect[0]*table[rarg]->vect[0] -
		    table[larg]->vect[1]*table[rarg]->vect[1] -
		    table[larg]->vect[2]*table[rarg]->vect[2] -
		    table[larg]->vect[3]*table[rarg]->vect[3];

 		curr->vect[1] = 
		    table[larg]->vect[1]*table[rarg]->vect[0] +	// maybe 
		    table[larg]->vect[0]*table[rarg]->vect[1] +	// optimize
		    table[larg]->vect[2]*table[rarg]->vect[3] -	// this
		    table[larg]->vect[3]*table[rarg]->vect[2];

 		curr->vect[2] = 
		    table[larg]->vect[2]*table[rarg]->vect[0] +
		    table[larg]->vect[0]*table[rarg]->vect[2] +
		    table[larg]->vect[3]*table[rarg]->vect[1] -
		    table[larg]->vect[1]*table[rarg]->vect[3];

 		curr->vect[3] = 
		    table[larg]->vect[3]*table[rarg]->vect[0] +
		    table[larg]->vect[0]*table[rarg]->vect[3] +
		    table[larg]->vect[1]*table[rarg]->vect[2] -
		    table[larg]->vect[2]*table[rarg]->vect[1];
		break;

	    case Q_TIMES_S:	// q*s
		curr->vect[0] = table[rarg]->vect[0]*table[larg]->vect[0];
		curr->vect[1] = table[rarg]->vect[0]*table[larg]->vect[1];
		curr->vect[2] = table[rarg]->vect[0]*table[larg]->vect[2];
		curr->vect[3] = table[rarg]->vect[0]*table[larg]->vect[3];
#ifdef OUTPUT
		printf("\tq*s : %.2f * %.2f is %.2f\n", table[larg]->vect[1],
		       table[rarg]->vect[0], curr->vect[1]);
#endif
		break;
				
	    case S_TIMES_S:	// s*s
		curr->vect[0] = table[larg]->vect[0]*table[rarg]->vect[0];
#ifdef OUTPUT
		printf("\t%.2f (%d) * %.2f (%d) is %.2f\n", 
		       table[larg]->vect[0], larg,
		       table[rarg]->vect[0], rarg, 
		       curr->vect[0]);
#endif
		break;

	    case Q_DIV_S:	// q/s
		curr->vect[0] = table[larg]->vect[0]/table[rarg]->vect[0];
		curr->vect[1] = table[larg]->vect[1]/table[rarg]->vect[0];
		curr->vect[2] = table[larg]->vect[2]/table[rarg]->vect[0];
		curr->vect[3] = table[larg]->vect[3]/table[rarg]->vect[0];
		break;
		
	    case S_DIV_S:		// s/s
		curr->vect[0] = table[larg]->vect[0]/table[rarg]->vect[0];
		break;

	    case Q_EXP_S_EVEN:	// (x^y), y even (quaternion ^ scalar)
		curr->vect[0] = 
		    table[larg]->vect[0]*table[larg]->vect[0] -
		    table[larg]->vect[1]*table[larg]->vect[1] -
		    table[larg]->vect[2]*table[larg]->vect[2] -
		    table[larg]->vect[3]*table[larg]->vect[3];
		
		curr->vect[1] = 2*table[larg]->vect[1]*table[larg]->vect[0];
		curr->vect[2] = 2*table[larg]->vect[2]*table[larg]->vect[0];
		curr->vect[3] = 2*table[larg]->vect[3]*table[larg]->vect[0];
		break;

	    case S_EXP_Q:	// (s^q) (scalar ^ quaternion)
		log_of_base = log(table[larg]->vect[0]);
		curr->vect[0] =
		    cos(log_of_base*table[rarg]->vect[1])*
		    cos(log_of_base*table[rarg]->vect[2])*
		    cos(log_of_base*table[rarg]->vect[3]) - 
		    sin(log_of_base*table[rarg]->vect[1])*
		    sin(log_of_base*table[rarg]->vect[2])*
		    sin(log_of_base*table[rarg]->vect[3]);

		curr->vect[1] =
		    cos(log_of_base*table[rarg]->vect[1])*
		    sin(log_of_base*table[rarg]->vect[2])*
		    sin(log_of_base*table[rarg]->vect[3]) +
		    sin(log_of_base*table[rarg]->vect[1])*
		    cos(log_of_base*table[rarg]->vect[2])*
		    cos(log_of_base*table[rarg]->vect[3]);

		curr->vect[2] =
		    cos(log_of_base*table[rarg]->vect[1])*
		    sin(log_of_base*table[rarg]->vect[2])*
		    cos(log_of_base*table[rarg]->vect[3]) - 
		    sin(log_of_base*table[rarg]->vect[1])*
		    cos(log_of_base*table[rarg]->vect[2])*
		    sin(log_of_base*table[rarg]->vect[3]);

		curr->vect[3] =
		    cos(log_of_base*table[rarg]->vect[1])*
		    cos(log_of_base*table[rarg]->vect[2])*
		    sin(log_of_base*table[rarg]->vect[3]) + 
		    sin(log_of_base*table[rarg]->vect[1])*
		    sin(log_of_base*table[rarg]->vect[2])*
		    cos(log_of_base*table[rarg]->vect[3]);
		break;
		
	    case S_EXP_S:
		curr->vect[0] = pow(table[larg]->vect[0],table[rarg]->vect[0]);
		break;

	    case RE_Q:		// re(q)
		curr->vect[0] = table[larg]->vect[0];
#ifdef OUTPUT
		printf("\tre(%.2f) is %.2f\n", table[larg]->vect[0],
		       curr->vect[0]);
#endif
		break;
		
	    case IM_I_Q:	// im_i(q)
		curr->vect[0] = table[larg]->vect[1];
		break;

	    case IM_J_Q:	// im_j(q)
		curr->vect[0] = table[larg]->vect[2];
		break;

	    case IM_K_Q:	// im_k(q)
		curr->vect[0] = table[larg]->vect[3];
		break;

	    case ABS:
		curr->vect[0] = fabs(table[larg]->vect[0]);
		printf("\tabs of %.2f is %.2f\n", 
		       table[larg]->vect[0], 
		       curr->vect[0]);
		break;
	    }

    }
}

void eq_table::set_init_q_val(double * v) {
  printf("don't call set_init_q_val\n");
  exit(-1);
    init_q_val[0] = v[0];
    init_q_val[1] = v[1];
    init_q_val[2] = v[2];
    init_q_val[3] = v[3];
    q_is_c_initially = false;
}

void eq_table::set_q_to_c_initially() {
    q_is_c_initially = true;
}
