%{
#include "quatdefs.h"
#define MATCH(s) return(match(s))
#ifndef YY_USE_PROTOS
#define YY_USE_PROTOS
#endif

%}


DIGIT [0-9]
FLOAT ({DIGIT}+("."{DIGIT}+)?)|("."{DIGIT}+)


%%

[*^/+(|)qzijkcen_-]         { MATCH(*yytext); }
pi                          { MATCH(TOK_LIT_PI); }
re                          { MATCH(TOK_RE); }
im_i                        { MATCH(TOK_IM_I); }
im_j                        { MATCH(TOK_IM_J); }
im_k                        { MATCH(TOK_IM_K); }
im                          { MATCH(TOK_IM_I); }
sin                         { MATCH(TOK_SIN); }
cos                         { MATCH(TOK_COS); }
tan                         { MATCH(TOK_TAN); }
sinh                        { MATCH(TOK_SINH); }
cosh                        { MATCH(TOK_COSH); }
tanh                        { MATCH(TOK_TANH); }
ln                          { MATCH(TOK_LN); }
log                         { MATCH(TOK_LOG); }
{FLOAT}                     { MATCH(TOK_LIT_SCALAR); }
\[" "*-?{FLOAT}" "*-?{FLOAT}" "*-?{FLOAT}" "*-?{FLOAT}" "*\]    { MATCH(TOK_LIT_QUAT); }
\[" "*-?{FLOAT}" "*-?{FLOAT}" "*\]    { MATCH(TOK_LIT_COMPLEX); }
[\n\t ]+                   {  }
.                          { char buf[1024] ; sprintf(buf, "unrecognized token '%s'\n", yytext) ; yyerror(buf); }

%%

int match(int token_code) {

    expr_ref ans = new Expr_Tree(yytext, token_code);
    double a, b, c, d;

    switch(token_code ) {

    case TOK_LIT_QUAT:
	sscanf(yytext, "[ %lf %lf %lf %lf ]", &a, &b, &c, &d);
	ans->setVal(new quat(a, b, c, d));
	ans->set_constant();
        break;

    case TOK_LIT_COMPLEX:
	sscanf(yytext, "[ %lf %lf ]", &a, &b, &c, &d);
	ans->setVal(new quat(a, b, 0, 0));
	ans->set_constant();
        ans->set_complex();
        break;

    case TOK_LIT_SCALAR:
	double s;
	sscanf(yytext, " %lf ", &s);
        ans->setVal(s);
	ans->set_scalar();
	ans->set_constant();
        //printf("found scalar - val = %.2f\n", ans->scalar_val);
        break;

    case 'e':
	ans->setVal(2.71828182846);
	ans->set_scalar();
	ans->set_constant();
        break;

    case TOK_LIT_PI:
	ans->setVal(3.1415926535897);
	ans->set_scalar();
	ans->set_constant();
        break;

    case 'i':
	ans->setVal(new quat(0, 1, 0, 0));
	ans->set_constant();
        ans->set_complex();
        break;

    case 'j':
	ans->setVal(new quat(0, 0, 1, 0));
	ans->set_constant();
        break;

    case 'k':
	ans->setVal(new quat(0, 0, 0, 1));
	ans->set_constant();
        break;

    case 'n':
        ans->set_scalar();
        break;

    case 'z':
        ans->set_complex();
        break;
    }

    yylval = ans;

    return (token_code);
}

int yywrap() {}





