/*
 *  expr_tree.cc
 * 
 *  Fractal Generator
 * 
 *  Copyright (c) 1998 Brian Martin
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "quatdefs.h"
#include <math.h>

char *Expr_Tree::make_string() {
    
    char buf[200];
    buf[0] = 0;
    char *ptr = 0;

    if(lchld) {
	ptr = lchld->make_string();
	strcat(buf, ptr);
	delete ptr;
    }

    strcat(buf, data);

    if(rchld) {
	ptr = rchld->make_string();
	strcat(buf, ptr);
	delete ptr;
    }
    return strdup(buf);
}

void Expr_Tree::dump() {
    
    if(lchld || rchld)
	printf(" ( ");
    
    if(lchld)
 	lchld->dump();

    if(val)
	val->dump();
    else 
	printf("%s", data);

    if(rchld)
 	rchld->dump();

    if(lchld || rchld)
	printf(" ) ");
}

expr_ref Expr_Tree::set_child(expr_ref child) {
    
    rchld = child;
    lchld = NULL;

    return this;
}

/*
  This routine attempts to fold constants as much as possible
 */
expr_ref Expr_Tree::set_children(expr_ref _lchld, expr_ref _rchld) {

    rchld = _rchld;
    lchld = _lchld;

//     printf("expr is ");
//     dump();
//     printf("\n");

    if(rchld->scalar && lchld->scalar && rchld->constant && lchld->constant) {
	scalar = true;
	constant = true;
	if(!val)
	    val = new quat((double)0);
	switch (token_code) {

	case '*':
	    val->s = rchld->val->s * lchld->val->s;
	    break;

	case '/':
	    val->s = lchld->val->s / rchld->val->s;
	    break;

	case '+':
	    val->s = rchld->val->s + lchld->val->s;
	    break;

	case '-':
	    val->s = lchld->val->s - rchld->val->s;
	    break;

	case '^':
	    val->s = pow(lchld->val->s, rchld->val->s);
	    break;
	    
	default:
	    fprintf(stderr, "Unrecognized operator %d\n", token_code);
	    exit(-1);
	    break;
	}
	token_code = TOK_LIT_SCALAR;
	delete rchld;
	delete lchld;
	lchld = rchld = NULL;
// 	printf("folded constants - %.2f\n", val->s);
    } else if (rchld->constant && lchld->constant) {
	constant = true;
	if(lchld->scalar) {

	    double log_base;
	    double base_s;

	    val = new quat(rchld->val);
	    
	    switch (token_code) {
	    case '*':
		val->s *= lchld->val->s;
		val->a *= lchld->val->s;
		val->b *= lchld->val->s;
		val->c *= lchld->val->s;
		break;

	    case '+':
		val->s += lchld->val->s;
		val->a += lchld->val->s;
		val->b += lchld->val->s;
		val->c += lchld->val->s;
		break;

	    case '-':
		val->s -= lchld->val->s;
		val->a -= lchld->val->s;
		val->b -= lchld->val->s;
		val->c -= lchld->val->s;
		break;

	    case '^':		// more difficult (this is wrong I think)
		log_base = log(lchld->val->s);
		base_s = pow(lchld->val->s, rchld->val->s);
		val->s = base_s * (cos(rchld->val->a*log_base)*
				   cos(rchld->val->b*log_base)*
				   cos(rchld->val->c*log_base) -
				   sin(rchld->val->a*log_base)*
				   sin(rchld->val->b*log_base)*
				   sin(rchld->val->c*log_base));
		val->a = base_s * (cos(rchld->val->a*log_base)*
				   sin(rchld->val->b*log_base)*
				   sin(rchld->val->c*log_base) +
				   sin(rchld->val->a*log_base)*
				   cos(rchld->val->b*log_base)*
				   cos(rchld->val->c*log_base));
		val->b = base_s * (cos(rchld->val->a*log_base)*
				   sin(rchld->val->b*log_base)*
				   cos(rchld->val->c*log_base) -
				   sin(rchld->val->a*log_base)*
				   cos(rchld->val->b*log_base)*
				   sin(rchld->val->c*log_base));
		val->c = base_s * (cos(rchld->val->a*log_base)*
				   cos(rchld->val->b*log_base)*
				   sin(rchld->val->c*log_base) +
				   sin(rchld->val->a*log_base)*
				   sin(rchld->val->b*log_base)*
				   cos(rchld->val->c*log_base));
		break;
		
	    default:
 		fprintf(stderr, "Unrecognized operator %d\n", token_code);
		exit(-1);
		break;

	    }
	    token_code = TOK_LIT_QUAT;
	} else if (rchld->scalar) {
	    val = new quat(lchld->val);
	    
	    switch (token_code) {
	    case '*':
		val->s *= rchld->val->s;
		val->a *= rchld->val->s;
		val->b *= rchld->val->s;
		val->c *= rchld->val->s;
		break;

	    case '/':
		val->s /= rchld->val->s;
		val->a /= rchld->val->s;
		val->b /= rchld->val->s;
		val->c /= rchld->val->s;
		break;

	    case '+':
		val->s += rchld->val->s;
		val->a += rchld->val->s;
		val->b += rchld->val->s;
		val->c += rchld->val->s;
		break;

	    case '-':
		val->s -= rchld->val->s;
		val->a -= rchld->val->s;
		val->b -= rchld->val->s;
		val->c -= rchld->val->s;
		break;

	    case '^':		// more difficult
				// umm...
		fprintf(stderr, "not implemented yet ;<\n");
		exit(-1);
		break;

	    default:
		fprintf(stderr, "Unrecognized operator %d\n", token_code);
		exit(-1);
		break;
	    }
	    token_code = TOK_LIT_QUAT;
	} else {		// no scalars involved
	    val = new quat;
	    switch (token_code) {
	    case '*':
		val->s = 
		    rchld->val->s*lchld->val->s -
		    rchld->val->a*lchld->val->a -
		    rchld->val->b*lchld->val->b -
		    rchld->val->c*lchld->val->c;
		val->a = 
		    lchld->val->a*rchld->val->s +
		    lchld->val->s*rchld->val->a +
		    lchld->val->b*rchld->val->c -
		    lchld->val->c*rchld->val->b;
		val->b = 
		    lchld->val->b*rchld->val->s +
		    lchld->val->s*rchld->val->b +
		    lchld->val->c*rchld->val->a -
		    lchld->val->a*rchld->val->c;
		val->c = 
		    lchld->val->c*rchld->val->s +
		    lchld->val->s*rchld->val->c +
		    lchld->val->a*rchld->val->b -
		    lchld->val->b*rchld->val->a;
		break;

	    case '+':
		val->s = rchld->val->s+lchld->val->s;
		val->a = rchld->val->a+lchld->val->a;
		val->b = rchld->val->b+lchld->val->b;
		val->c = rchld->val->c+lchld->val->c;
		break;

	    case '-':
		val->s = rchld->val->s-lchld->val->s;
		val->a = rchld->val->a-lchld->val->a;
		val->b = rchld->val->b-lchld->val->b;
		val->c = rchld->val->c-lchld->val->c;
		break;
		
	    default:
		fprintf(stderr, "Unrecognized operator %d\n", token_code);
		exit(-1);
		break;
	    }
	    token_code = TOK_LIT_QUAT;
	}
	delete rchld;
	delete lchld;
	lchld = rchld = NULL;
// 	printf("folded constants - %.2f %.2f %.2f %.2f\n", val->s, val->a, val->b, val->c);
    }

//     printf("after expr is ");
//     dump();
//     printf("\n");

    return(this);
}

Expr_Tree::Expr_Tree(char * _data, int _token_code) {

    data = strdup(_data);
    token_code = _token_code;
    
    lchld = rchld = 0;
    val = 0;
    scalar = false;
    constant = false;
    cross_product_drops_out = false;
}

Expr_Tree::~Expr_Tree() {

    if(val)
	delete val;
    if(data)
	delete data;

}

bool Expr_Tree::eql(Expr_Tree * comp) {
    
    return(this && comp && 
	   ((constant && 
	     ((val && comp->val && val->eql(comp->val)))) ||
	    (!val && !comp->val && strcmp(data, comp->data)==0)) &&
	    ((lchld && lchld->eql(comp->lchld)) || !lchld) &&
	    ((rchld && rchld->eql(comp->rchld)) || !rchld));
}

// expr_ref Expr_Tree::load(FILE * file) {
//     int count;
//     int token_code;
//     char buf[200];
//     expr_ref curr;
//     expr_ref old;
//     expr_ref ret_val;

//     fscanf(file, " %d ", &count);

//     fscanf(file, " %s %d ", buf, &token_code);
//     old = ret_val = new Expr_Tree(buf, token_code);

//     for(int i = 1 ; i < count ; i++) {
// 	fscanf(file, " %s %d ", buf, &token_code);
// 	curr = new Expr_Tree(buf, token_code);
// 	curr->rear = old;
// 	old->next = curr;
//     }
// }

// void Expr_Tree::save(FILE * file) {
//     int count = 0;
//     for(expr_ref curr = this ; curr ; curr = curr->next) count++;
//     fprintf(file, "%d\n", count);
//     for(expr_ref curr = this ; curr ; curr = curr->next) {
// 	fprintf(file, "%s\n %d\n", curr->data, curr->token_code);
// 	if(curr->val)
// 	    curr->val->save(file);
//     }
// }

Expr_Tree * Expr_Tree::copy() {
    
    Expr_Tree * copy = new Expr_Tree(data,token_code);
    copy->scalar = scalar;
    copy->constant = constant;
    copy->rank = rank;
    
    if(lchld)
      copy->lchld = lchld->copy();
    if(rchld)
      copy->rchld = rchld->copy();
    if(val)
      copy->val = new quat(val);
    return copy;
}

Expr_Tree * Expr_Tree::post_process() {

    deal_with_exponents();
    int type = get_eq_type();	// complex or quaternion
    assign_retvals(type);	// determine context of nodes -
				// [scalar|complex|quaternion]

    schedule();			// rank nodes for scheduling
    return this;
}

int Expr_Tree::get_eq_type() {
    
    int retval = -1;
    int child1 = -1;
    int child2 = -1;

    if(lchld)
	child1 = lchld->get_eq_type();
    if(rchld)
	child2 = rchld->get_eq_type();

    if(child1 != -1) retval = child1;
    if(child2 != -1) retval = child2;
    
    if(retval == -1)    
	if(token_code == 'z')
	    retval = 1;
	else if(token_code == 'q') 
	    retval = 0;

    return retval;
}


// this routine expresses all quaternion exponentiation in terms of squaring
// and multiplication
void Expr_Tree::deal_with_exponents() {

    char buf[50];

    if(lchld)
	lchld->deal_with_exponents();
    if(rchld)
	rchld->deal_with_exponents();
    
    if(lchld && !lchld->scalar && token_code == '^') {
        if(((int)rchld->val->s)%2) { // odd exponent
	    if(((int)rchld->val->s) == 1) {
		delete rchld;
		data = strdup(lchld->data);
		token_code = lchld->token_code;
		if(val)
		    delete val;
		if(lchld->val)
		    val = new quat(lchld->val);
		delete lchld;
		rchld = lchld = NULL;
	    } else {
		expr_ref new_child = copy();
		sprintf(buf, "%d", int(rchld->val->s)-1);
		new_child->rchld->data = strdup(buf);
		new_child->rchld->val->s = rchld->val->s -1;
		token_code = '*';
		cross_product_drops_out = true;
		sprintf(buf, "%c", '*');
		delete data;
		data = strdup(buf);
		delete rchld;
		rchld = new_child;
		rchld->deal_with_exponents();
	    } 
	} else {
	    if(rchld->val->s != 2) {
		expr_ref new_child = copy();
		int power = (int)((new_child->rchld->val->s)/2);
		sprintf(buf, "%d", power);
		new_child->rchld->data = strdup(buf);
		new_child->rchld->val->s = power;
		sprintf(buf, "%d", 2);
		rchld->data = strdup(buf);
		rchld->val->s = 2;
		delete lchld;
		lchld = new_child;
		lchld->deal_with_exponents();
	    }
	}
    }
}

void Expr_Tree::schedule() {

    rank = -1;
    
    if(token_code != '_') {
	if(lchld) {
	    lchld->schedule();
	    rank = lchld->rank;
	}
	if(rchld) {
	    rchld->schedule();
	    if(rank < rchld->rank)
		rank = rchld->rank;
	}
    }
    
    rank++;
}

void Expr_Tree::force_complex() {

    if(lchld)
	lchld->force_complex();
    if(rchld)
	rchld->force_complex();

    if(!scalar && !complex)
	complex = true;
}

void Expr_Tree::assign_retvals(int type) {

    if(lchld)
	lchld->assign_retvals(type);
    if(rchld)
	rchld->assign_retvals(type);

				// these operators force scalar context
    if(rchld || lchld) {
	scalar = false;
	if(token_code == TOK_RE || token_code == TOK_IM_I ||
	   token_code == TOK_IM_J || token_code == TOK_IM_K ||
	   token_code == '|')
	    scalar = true;
	else
	    if(rchld) {		// others depend upon context of operands
		if(lchld) {
		    if(lchld->scalar && rchld->scalar) 
			scalar = true;
		} else if(rchld->scalar)
		    scalar = true;
	    } else if(lchld && lchld->scalar) 
		scalar = true;

	complex = false;
	if(rchld) {		
	    if(lchld) {
		if((lchld->complex && rchld->complex) || 
		   (lchld->complex && rchld->scalar) ||
		   (lchld->scalar && rchld->complex))
		    complex = true;
	    } else if(rchld->complex)
		complex = true;
	} else if(lchld && lchld->complex) 
	    complex = true;
    } else {
	if(token_code == 'c' || token_code == 'i') {
	    complex = type ? true : false;
	} else if(token_code == 'z') { // this is a hack.  on z^4+c complex
	    complex = true;	// wasn't being recgonized properly
	} else if(!type) {
	    complex = false;
	}
    }
#if 0
    printf("assigning complex = %d to %s", complex, data);
     if(lchld) printf(" lchld->complex = %d", lchld->complex);
     if(rchld) printf(" rchld->complex = %d", rchld->complex);
     printf("\n");
     if(lchld && rchld)
	 printf("\tlchld = %s rchild = %s\n", lchld->data, rchld->data);
#endif
}

